from argparse import ArgumentParser
from statistics import mean
from sys import getsizeof

from humanize import naturalsize

from femos.core import get_number_of_nn_weights, get_evolved_population, Summary, get_end_datetime
from femos.genotypes import SimpleGenotype, UncorrelatedOneStepSizeGenotype, UncorrelatedNStepSizeGenotype
from femos.phenotypes import Phenotype
from femos.selections import get_two_size_tournament_parent_selection, get_n_size_tournament_parent_selection, \
    get_age_based_offspring_selection

# Available genotype choices
simple_genotype_choice = 'simple_genotype'
uncorrelated_one_step_size_genotype_choice = 'uoss_genotype'
uncorrelated_n_step_size_genotype_choice = 'unss_genotype'

genotype_lookup = {
    simple_genotype_choice: 'Simple genotype (simple_genotype)',
    uncorrelated_one_step_size_genotype_choice: 'Uncorrelated one step size genotype (uoss_genotype)',
    uncorrelated_n_step_size_genotype_choice: 'Uncorrelated n step size genotype (unss_genotype)',
}

# Available summary choices
summary_epoch_choice = 'epoch'
summary_mean_choice = 'mean'
summary_stddev_choice = 'stddev'
summary_population_size_choice = 'population_size'
summary_duration_choice = 'duration'
summary_choices = [summary_epoch_choice, summary_mean_choice, summary_stddev_choice, summary_population_size_choice,
                   summary_duration_choice]

summary_lookup = {
    summary_epoch_choice: Summary.EPOCH,
    summary_mean_choice: Summary.MEAN,
    summary_stddev_choice: Summary.STDDEV,
    summary_population_size_choice: Summary.POPULATION_SIZE,
    summary_duration_choice: Summary.DURATION
}


def get_evolution_summary(arguments, input_nodes, output_nodes, memory_consumption_probe=100):
    output = [str.format('# Basic summary'),
              str.format('Input nodes: {}', input_nodes),
              str.format('Iutput nodes: {}', output_nodes),
              str.format('Genotype: {}', genotype_lookup[arguments.genotype]),
              str.format('Hidden layer nodes: {}', arguments.hidden_layer_nodes),
              str.format('Weight lower threshold: {}', arguments.weight_lower_threshold),
              str.format('Weight upper threshold: {}', arguments.weight_upper_threshold),
              str.format('Population size: {}', arguments.population_size),
              str.format('Tournament size: {}', arguments.tournament_size),
              str.format('Duration (hours): {}', arguments.duration),
              str.format('Use bias: {}', arguments.bias),
              str.format('# Genotype specific summary')]

    if arguments.genotype == simple_genotype_choice:
        output.append(str.format('Mutation mean: {}', arguments.mutation_mean))
        output.append(str.format('Mutation standard deviation: {}', arguments.mutation_standard_deviation))

    if arguments.genotype == uncorrelated_one_step_size_genotype_choice or arguments.genotype == uncorrelated_n_step_size_genotype_choice:
        output.append(
            str.format('Mutation step size lower threshold: {}', arguments.mutation_step_size_lower_threshold))
        output.append(
            str.format('Mutation step size upper threshold: {}', arguments.mutation_step_size_upper_threshold))
        output.append(str.format('Tau 1: {}', arguments.tau1))

        if arguments.genotype == uncorrelated_n_step_size_genotype_choice:
            output.append(str.format('Tau 2: {}', arguments.tau2))

    output.append(str.format('# Calculated summary'))
    number_of_nn_weights = get_number_of_nn_weights(input_nodes, arguments.hidden_layer_nodes,
                                                    output_nodes)
    output.append(str.format('Number of neural network weights: {}', number_of_nn_weights))

    # Calculating memory consumption
    demo_genotype_iterator = range(memory_consumption_probe)

    if arguments.genotype == simple_genotype_choice:
        demo_genotypes = map(
            lambda index: SimpleGenotype.get_random_genotype(number_of_nn_weights, arguments.weight_lower_threshold,
                                                             arguments.weight_upper_threshold), demo_genotype_iterator)

    if arguments.genotype == uncorrelated_one_step_size_genotype_choice:
        demo_genotypes = map(lambda index: UncorrelatedOneStepSizeGenotype.get_random_genotype(number_of_nn_weights,
                                                                                               arguments.weight_lower_threshold,
                                                                                               arguments.weight_upper_threshold,
                                                                                               arguments.mutation_step_size_lower_threshold,
                                                                                               arguments.mutation_step_size_upper_threshold),
                             demo_genotype_iterator)

    if arguments.genotype == uncorrelated_n_step_size_genotype_choice:
        demo_genotypes = map(lambda index: UncorrelatedNStepSizeGenotype.get_random_genotype(number_of_nn_weights,
                                                                                             arguments.weight_lower_threshold,
                                                                                             arguments.weight_upper_threshold,
                                                                                             arguments.mutation_step_size_lower_threshold,
                                                                                             arguments.mutation_step_size_upper_threshold),
                             demo_genotype_iterator)

    demo_genotype_sizes = list(map(lambda demo_genotype: getsizeof(demo_genotype.weights), demo_genotypes))
    mean_demo_genotype_size = mean(demo_genotype_sizes) * arguments.population_size
    output.append(str.format('Calculated memory consumption (Python list): {}', naturalsize(mean_demo_genotype_size)))
    output.append(str.format('Approximate end time: {}', get_end_datetime(arguments.duration).isoformat(sep=' ')))

    output.append(str.format('# Utils summary'))
    output.append(str.format('Epoch summary: {}', arguments.epoch_summary))
    if arguments.epoch_summary:
        output.append(str.format('Epoch summary features: {}', arguments.epoch_summary_features))
        output.append(str.format('Epoch summary interval: {}', arguments.epoch_summary_interval))

    output.append(str.format('Population backup summary: {}', arguments.population_backup))
    if arguments.population_backup:
        output.append(str.format('Population backup directory: {}', arguments.population_backup_directory))
        output.append(str.format('Population backup interval: {}', arguments.population_backup_interval))
        output.append(str.format('Population backup file extension: .{}', arguments.population_backup_file_extension))

    return '\n'.join(output)


def get_core_argument_parser():
    parser = ArgumentParser(description='Femos CLI - Command line interface for my small library for neuroevolution.')
    parser.add_argument('genotype', metavar='G', type=str,
                        choices=[simple_genotype_choice, uncorrelated_one_step_size_genotype_choice,
                                 uncorrelated_n_step_size_genotype_choice],
                        help='Describes genotype type: simple_genotype - SimpleGenotype, uoss_genotype - UncorrelatedOneStepSizeGenotype, '
                             'unss_genotype - UncorrelatedNStepSizeGenotype')

    parser.add_argument('--bias', help='Use bias', action='store_true', default=False)
    parser.add_argument('--hidden_layer_nodes', help='List of nodes in hidden layer', type=int, default=[],
                        nargs='+')
    parser.add_argument("--weight_lower_threshold", help="Lower threshold value of ann weights", type=float,
                        default=-1.0)
    parser.add_argument("--weight_upper_threshold", help="Upper threshold value of ann weights", type=float,
                        default=1.0)
    parser.add_argument("--mutation_step_size_lower_threshold",
                        help="Lower threshold value for genotype mutation step size",
                        type=float, default=-0.2)
    parser.add_argument("--mutation_step_size_upper_threshold",
                        help="Upper threshold value for genotype mutation step size",
                        type=float, default=0.2)
    parser.add_argument("--population_size", help="Number of individuals in population", type=int, default=20)
    parser.add_argument("--tournament_size", help="Number of individuals to rival in tournament selection",
                        type=int, default=3)
    parser.add_argument('--duration', help="Duration of evolution in hours", type=float, default=1)
    parser.add_argument("--tau1", help="Mutation operator parameter - tau1", type=float, default=0.001)
    parser.add_argument("--tau2", help="Mutation operator parameter - tau2", type=float, default=0.01)

    parser.add_argument('--mutation_mean', type=float, default=0)
    parser.add_argument('--mutation_standard_deviation', type=float, default=0.1)

    parser.add_argument('--epoch_summary', action='store_true', default=False)
    parser.add_argument('--epoch_summary_features', type=str, nargs='+', choices=summary_choices,
                        default=[summary_epoch_choice, summary_mean_choice, summary_stddev_choice,
                                 summary_duration_choice])
    parser.add_argument('--epoch_summary_interval', type=int, default=5)

    parser.add_argument('--population_backup', action='store_true', default=False)
    parser.add_argument('--population_backup_directory', type=str, default='backups')
    parser.add_argument('--population_backup_interval', type=int, default=5)
    parser.add_argument('--population_backup_file_extension', type=str, default='.population')
    parser.add_argument('--dry_run', action='store_true', default=False)
    return parser


def handle_evolution_run(input_nodes, output_nodes, evaluation_strategy):
    argument_parser = get_core_argument_parser()
    arguments = argument_parser.parse_args()
    evolution_summary = get_evolution_summary(arguments, input_nodes, output_nodes)

    print(evolution_summary)
    if not arguments.dry_run:

        def phenotype_strategy(genotype):
            return Phenotype.get_phenotype_from_genotype(genotype, input_nodes, arguments.hidden_layer_nodes,
                                                         output_nodes, arguments.bias)

        def parent_selection_strategy(phenotype_values):
            if arguments.tournament_size == 2:
                return get_two_size_tournament_parent_selection(phenotype_values, arguments.population_size)
            else:
                return get_n_size_tournament_parent_selection(phenotype_values, arguments.tournament_size,
                                                              arguments.population_size)

        def mutation_strategy(genotype):
            if arguments.genotype == simple_genotype_choice:
                return SimpleGenotype.get_mutated_genotype(genotype, arguments.mutation_mean,
                                                           arguments.mutation_standard_deviation)

            if arguments.genotype == uncorrelated_one_step_size_genotype_choice:
                return UncorrelatedOneStepSizeGenotype.get_mutated_genotype(genotype, arguments.tau1)

            if arguments.genotype == uncorrelated_n_step_size_genotype_choice:
                return UncorrelatedNStepSizeGenotype.get_mutated_genotype(genotype, arguments.tau1, arguments.tau2)

        def offspring_selection_strategy(parents, mutated_parents):
            return get_age_based_offspring_selection(parents, mutated_parents)

        number_of_nn_weights = get_number_of_nn_weights(input_nodes, arguments.hidden_layer_nodes,
                                                        output_nodes, arguments.bias)
        if arguments.genotype == simple_genotype_choice:
            initial_population = SimpleGenotype.get_random_genotypes(arguments.population_size, number_of_nn_weights,
                                                                     arguments.weight_lower_threshold,
                                                                     arguments.weight_upper_threshold)

        if arguments.genotype == uncorrelated_one_step_size_genotype_choice:
            initial_population = UncorrelatedOneStepSizeGenotype.get_random_genotypes(arguments.population_size,
                                                                                      number_of_nn_weights,
                                                                                      arguments.weight_lower_threshold,
                                                                                      arguments.weight_upper_threshold,
                                                                                      arguments.mutation_step_size_lower_threshold,
                                                                                      arguments.mutation_step_size_upper_threshold)

        if arguments.genotype == uncorrelated_n_step_size_genotype_choice:
            initial_population = UncorrelatedNStepSizeGenotype.get_random_genotypes(arguments.population_size,
                                                                                    number_of_nn_weights,
                                                                                    arguments.weight_lower_threshold,
                                                                                    arguments.weight_upper_threshold,
                                                                                    arguments.mutation_step_size_lower_threshold,
                                                                                    arguments.mutation_step_size_upper_threshold)

        population_backup = None
        if arguments.population_backup:
            population_backup = [arguments.population_backup_interval, arguments.population_backup_directory,
                                 arguments.population_backup_file_extension]

        epoch_summary_strategy = None
        if arguments.epoch_summary:
            summary_features = list(
                map(lambda summary_choice: summary_lookup[summary_choice], arguments.epoch_summary_features))
            epoch_summary_strategy = [summary_features, arguments.epoch_summary_interval]

        return get_evolved_population(initial_population, phenotype_strategy, evaluation_strategy,
                                      parent_selection_strategy, mutation_strategy, offspring_selection_strategy,
                                      arguments.duration, population_backup, epoch_summary_strategy)



from math import exp
from random import uniform, gauss

from .core import get_random_numbers


class SimpleGenotype:

    def __init__(self, weights):
        self.weights = weights

    @staticmethod
    def get_random_genotype(number_of_nn_weights, weight_lower_threshold, weight_upper_threshold):
        weights = get_random_numbers(number_of_nn_weights, weight_lower_threshold, weight_upper_threshold)
        return SimpleGenotype(weights)

    @staticmethod
    def get_random_genotypes(number_of_genotypes, number_of_nn_weights, weight_lower_threshold, weight_upper_threshold):
        genotypes = []
        for index in range(number_of_genotypes):
            genotype = SimpleGenotype.get_random_genotype(number_of_nn_weights, weight_lower_threshold,
                                                          weight_upper_threshold)
            genotypes.append(genotype)

        return genotypes

    @staticmethod
    def get_mutated_genotype(genotype, mean, standard_deviation):
        number_of_nn_weights = len(genotype.weights)

        new_weights = [0] * number_of_nn_weights
        for index in range(number_of_nn_weights):
            new_weights[index] = genotype.weights[index] + gauss(mean, standard_deviation)

        return SimpleGenotype(new_weights)

    def __eq__(self, other):
        return self.weights == other.weights


class UncorrelatedOneStepSizeGenotype:

    def __init__(self, weights, mutation_step_size):
        self.weights = weights
        self.mutation_step_size = mutation_step_size

    @staticmethod
    def get_random_genotype(number_of_nn_weights, weight_lower_threshold, weight_upper_threshold,
                            mutation_step_size_lower_threshold, mutation_step_size_upper_threshold):
        weights = get_random_numbers(number_of_nn_weights, weight_lower_threshold, weight_upper_threshold)
        mutation_step_size = uniform(mutation_step_size_lower_threshold, mutation_step_size_upper_threshold)
        return UncorrelatedOneStepSizeGenotype(weights, mutation_step_size)

    @staticmethod
    def get_random_genotypes(number_of_genotypes, number_of_nn_weights, weight_lower_threshold, weight_upper_threshold,
                             mutation_step_size_lower_threshold, mutation_step_size_upper_threshold):
        genotypes = []
        for index in range(number_of_genotypes):
            genotype = UncorrelatedOneStepSizeGenotype.get_random_genotype(number_of_nn_weights, weight_lower_threshold,
                                                                           weight_upper_threshold,
                                                                           mutation_step_size_lower_threshold,
                                                                           mutation_step_size_upper_threshold)
            genotypes.append(genotype)

        return genotypes

    @staticmethod
    def get_mutated_genotype(genotype, tau1):
        number_of_nn_weights = len(genotype.weights)

        # First mutate current mutation step size
        new_mutation_step_size = genotype.mutation_step_size * exp(tau1 * gauss(0, 1))

        # Then mutate neural network weights
        new_weights = [0] * number_of_nn_weights
        for index in range(number_of_nn_weights):
            new_weights[index] = genotype.weights[index] + new_mutation_step_size * gauss(0, 1)

        return UncorrelatedOneStepSizeGenotype(new_weights, new_mutation_step_size)

    def __eq__(self, other):
        return self.weights == other.weights and self.mutation_step_size == other.mutation_step_size


class UncorrelatedNStepSizeGenotype:

    def __init__(self, weights, mutation_step_sizes):
        self.weights = weights
        self.mutation_step_sizes = mutation_step_sizes

    @staticmethod
    def get_random_genotype(number_of_nn_weights, weight_lower_threshold, weight_upper_threshold,
                            mutation_step_size_lower_threshold, mutation_step_size_upper_threshold):
        weights = get_random_numbers(number_of_nn_weights, weight_lower_threshold, weight_upper_threshold)
        mutation_step_sizes = get_random_numbers(number_of_nn_weights, mutation_step_size_lower_threshold,
                                                 mutation_step_size_upper_threshold)
        return UncorrelatedNStepSizeGenotype(weights, mutation_step_sizes)

    @staticmethod
    def get_random_genotypes(number_of_genotypes, number_of_nn_weights, weight_lower_threshold, weight_upper_threshold,
                             mutation_step_size_lower_threshold, mutation_step_size_upper_threshold):
        genotypes = []
        for index in range(number_of_genotypes):
            genotype = UncorrelatedNStepSizeGenotype.get_random_genotype(number_of_nn_weights, weight_lower_threshold,
                                                                         weight_upper_threshold,
                                                                         mutation_step_size_lower_threshold,
                                                                         mutation_step_size_upper_threshold)
            genotypes.append(genotype)

        return genotypes

    @staticmethod
    def get_mutated_genotype(genotype, tau1, tau2):
        number_of_nn_weights = len(genotype.weights)

        # First mutate current mutation step sizes
        nonuniform_random_number = gauss(0, 1)
        new_mutation_step_sizes = [0] * number_of_nn_weights
        new_weights = [0] * number_of_nn_weights

        for index in range(number_of_nn_weights):
            new_mutation_step_sizes[index] = genotype.mutation_step_sizes[index] * exp(
                tau1 * nonuniform_random_number + tau2 * gauss(0, 1))
            new_weights[index] = genotype.weights[index] + new_mutation_step_sizes[index] * gauss(0, 1)

        return UncorrelatedNStepSizeGenotype(new_weights, new_mutation_step_sizes)

    def __eq__(self, other):
        return self.weights == other.weights and self.mutation_step_sizes == other.mutation_step_sizes


class CorrelatedNStepSizeGenotype:
    pass




from numpy import ndarray, array

from femos.core import get_number_of_nn_weights
from femos.genotypes import SimpleGenotype
from femos.phenotypes import Phenotype


def test_phenotype_get_phenotype_for_genotype():
    input_nodes = 4
    hidden_layers_nodes = [6, 8]
    output_nodes = 3

    number_of_nn_weights = get_number_of_nn_weights(input_nodes, hidden_layers_nodes, output_nodes)
    weight_lower_threshold = -1
    weight_upper_threshold = 1

    random_simple_genotype = SimpleGenotype.get_random_genotype(number_of_nn_weights, weight_lower_threshold,
                                                                weight_upper_threshold)

    phenotype = Phenotype.get_phenotype_from_genotype(random_simple_genotype, input_nodes, hidden_layers_nodes,
                                                      output_nodes)

    assert type(phenotype) is Phenotype
    assert len(phenotype.layers) == 3
    assert phenotype.input_nodes == input_nodes
    assert phenotype.hidden_layers_nodes == hidden_layers_nodes
    assert phenotype.output_nodes == output_nodes

    for element in phenotype.layers:
        assert type(element) is ndarray


def test_phenotype_get_prediction():
    input_nodes = 4
    hidden_layers_nodes = [6, 8]
    output_nodes = 3

    number_of_nn_weights = get_number_of_nn_weights(input_nodes, hidden_layers_nodes, output_nodes)
    weight_lower_threshold = -1
    weight_upper_threshold = 1

    random_simple_genotype = SimpleGenotype.get_random_genotype(number_of_nn_weights, weight_lower_threshold,
                                                                weight_upper_threshold)

    phenotype = Phenotype.get_phenotype_from_genotype(random_simple_genotype, input_nodes, hidden_layers_nodes,
                                                      output_nodes)

    input_values = array([[1, 0, 0, -1]])

    prediction = Phenotype.get_prediction(phenotype, input_values)

    assert type(prediction) is ndarray
    assert prediction.shape == (1, 3)



from random import random

from femos.selections import get_two_size_tournament_parent_selection, get_n_size_tournament_parent_selection


def test_two_size_tournament_parent_selection():
    number_of_individuals = 30
    individual_values = [0] * 30

    for index in range(number_of_individuals):
        individual_values[index] = random()

    selected_parents = get_two_size_tournament_parent_selection(individual_values, number_of_individuals)

    assert len(selected_parents) == number_of_individuals
    for element in selected_parents:
        assert type(element) is int
        assert 0 <= element <= (number_of_individuals - 1)


def test_n_size_tournament_selection():
    number_of_individuals = 100
    individual_values = [0] * number_of_individuals

    for index in range(number_of_individuals):
        individual_values[index] = random()

    selected_parents = get_n_size_tournament_parent_selection(individual_values, 5, number_of_individuals)
    assert len(selected_parents) == number_of_individuals
    for element in selected_parents:
        assert type(element) is int
        assert 0 <= element <= (number_of_individuals - 1)



from random import randint, sample


def get_two_size_tournament_parent_selection(individual_values, number_of_individuals):
    indices = [0] * number_of_individuals

    for index in range(number_of_individuals):
        first_individual_index = randint(0, number_of_individuals - 1)
        second_individual_index = randint(0, number_of_individuals - 1)

        first_individual_value = individual_values[first_individual_index]
        second_individual_value = individual_values[second_individual_index]

        if first_individual_value >= second_individual_value:
            indices[index] = first_individual_index
        else:
            indices[index] = second_individual_index

    return indices


def get_n_size_tournament_parent_selection(individual_values, tournament_size, number_of_individuals):
    indices = [0] * number_of_individuals
    individuals_indices = list(range(number_of_individuals))

    for index in range(number_of_individuals):
        selected_individuals = sample(individuals_indices, tournament_size)
        selected_individual_values = map(lambda individual_index: individual_values[individual_index],
                                         selected_individuals)

        zipped_values_with_indices = zip(selected_individual_values, selected_individuals)
        indices[index] = max(list(zipped_values_with_indices))[1]

    return indices


def get_age_based_offspring_selection(parents, updated_parents):
    return updated_parents



from numpy import array, matmul, tanh


class Phenotype:

    def __init__(self, weights, input_nodes, hidden_layers_nodes, output_nodes, use_bias):
        self.layers = []
        self.bias_layers = []

        grouped_nodes = [input_nodes] + hidden_layers_nodes + [output_nodes]
        weights_offset = 0

        for index in range(len(grouped_nodes) - 1):
            number_of_weights_per_layer = grouped_nodes[index] * grouped_nodes[index + 1]
            weights_per_layer = weights[weights_offset:weights_offset + number_of_weights_per_layer]
            weights_offset += number_of_weights_per_layer

            layer_matrix = array(weights_per_layer)
            layer_matrix = layer_matrix.reshape(grouped_nodes[index], grouped_nodes[index + 1])
            self.layers.append(layer_matrix)

        if use_bias:
            number_of_bias_weights = sum(hidden_layers_nodes) + output_nodes
            bias_weights = weights[weights_offset:weights_offset + number_of_bias_weights]

            bias_weights_offset = 0
            for units in grouped_nodes[1:]:
                selected_bias_weights = bias_weights[bias_weights_offset:bias_weights_offset + units]
                bias_weights_offset += units
                bias_matrix = array(selected_bias_weights).reshape(1, units)
                self.bias_layers.append(bias_matrix)

        self.input_nodes = input_nodes
        self.hidden_layers_nodes = hidden_layers_nodes
        self.output_nodes = output_nodes
        self.use_bias = use_bias

    @staticmethod
    def get_phenotype_from_genotype(genotype, input_nodes, hidden_layer_nodes, output_nodes, bias_weights=False):
        weights = genotype.weights
        return Phenotype(weights, input_nodes, hidden_layer_nodes, output_nodes, bias_weights)

    @staticmethod
    def get_prediction(phenotype, input_values, activation_function=tanh):
        number_of_layers = len(phenotype.hidden_layers_nodes) + 1

        result = input_values
        if phenotype.use_bias:
            for index in range(number_of_layers):
                result = activation_function(matmul(result, phenotype.layers[index]) + phenotype.bias_layers[index])

            return result
        else:
            for index in range(number_of_layers):
                result = activation_function(matmul(result, phenotype.layers[index]))

            return result



from glob import glob
from random import random
from shutil import rmtree

from femos.core import get_number_of_nn_weights, get_random_numbers, get_next_population, get_population_file_name, \
    get_evolved_population, handle_backup_load
from femos.genotypes import SimpleGenotype
from femos.phenotypes import Phenotype
from femos.selections import get_two_size_tournament_parent_selection, get_age_based_offspring_selection


def test_get_random_numbers():
    quantity = 500
    lower_threshold = -1
    upper_threshold = 1

    random_numbers = get_random_numbers(quantity, lower_threshold, upper_threshold)
    assert len(random_numbers) == quantity

    for element in random_numbers:
        assert lower_threshold <= element <= upper_threshold


def test_get_number_of_nn_weights():
    assert get_number_of_nn_weights(9, [4, 4], 9) == 88
    assert get_number_of_nn_weights(9, [], 9) == 81
    assert get_number_of_nn_weights(9, [8, 8, 8], 1) == 208


def test_get_next_population():
    number_of_genotypes = 40
    input_nodes = 4
    hidden_layers_nodes = []
    output_nodes = 3
    number_of_nn_weights = get_number_of_nn_weights(input_nodes, hidden_layers_nodes, output_nodes)
    weight_lower_threshold = -1
    weight_upper_threshold = 1
    mutation_mean = 0
    mutation_standard_deviation = 0.2

    def phenotype_strategy(genotype):
        return Phenotype.get_phenotype_from_genotype(genotype, input_nodes, hidden_layers_nodes, output_nodes)

    # Fake evaluation function, assign random number form 0 to 1 to each phenotype
    def evaluation_strategy(phenotypes):
        number_of_phenotypes = len(phenotypes)

        evaluation = [0] * number_of_phenotypes
        for index in range(number_of_phenotypes):
            evaluation[index] = random()

        return evaluation

    def parent_selection_strategy(phenotypes_values):
        return get_two_size_tournament_parent_selection(phenotypes_values, number_of_genotypes)

    def mutation_strategy(genotype):
        return SimpleGenotype.get_mutated_genotype(genotype, mutation_mean, mutation_standard_deviation)

    def offspring_selection_strategy(parents, mutated_parents):
        return get_age_based_offspring_selection(parents, mutated_parents)

    initial_population = SimpleGenotype.get_random_genotypes(number_of_genotypes, number_of_nn_weights,
                                                             weight_lower_threshold, weight_upper_threshold)

    next_population, phenotype_values, start_time, end_time = get_next_population(initial_population, phenotype_strategy, evaluation_strategy,
                                          parent_selection_strategy, mutation_strategy, offspring_selection_strategy)

    assert len(next_population) == number_of_genotypes

    for element in next_population:
        assert type(element) is SimpleGenotype


def test_get_population_file_name():
    epoch = 100

    generated_population_name = get_population_file_name(epoch, 4, ".popu")
    assert generated_population_name == "0100.popu"

    epoch = 20
    generated_population_name = get_population_file_name(epoch, 2, ".population")
    assert generated_population_name == "20.population"


def test_population_backup():
    number_of_genotypes = 40
    input_nodes = 4
    hidden_layers_nodes = []
    output_nodes = 3
    number_of_nn_weights = get_number_of_nn_weights(input_nodes, hidden_layers_nodes, output_nodes)
    weight_lower_threshold = -1
    weight_upper_threshold = 1
    mutation_mean = 0
    mutation_standard_deviation = 0.2

    def phenotype_strategy(genotype):
        return Phenotype.get_phenotype_from_genotype(genotype, input_nodes, hidden_layers_nodes, output_nodes)

    # Fake evaluation function, assign random number form 0 to 1 to each phenotype
    def evaluation_strategy(phenotypes):
        number_of_phenotypes = len(phenotypes)

        evaluation = [0] * number_of_phenotypes
        for index in range(number_of_phenotypes):
            evaluation[index] = random()

        return evaluation

    def parent_selection_strategy(phenotypes_values):
        return get_two_size_tournament_parent_selection(phenotypes_values, number_of_genotypes)

    def mutation_strategy(genotype):
        return SimpleGenotype.get_mutated_genotype(genotype, mutation_mean, mutation_standard_deviation)

    def offspring_selection_strategy(parents, mutated_parents):
        return get_age_based_offspring_selection(parents, mutated_parents)

    initial_population = SimpleGenotype.get_random_genotypes(number_of_genotypes, number_of_nn_weights,
                                                             weight_lower_threshold, weight_upper_threshold)

    backup = [5, "backups/evo1", ".population"]
    evolved_population = get_evolved_population(initial_population, phenotype_strategy, evaluation_strategy,
                                                parent_selection_strategy, mutation_strategy,
                                                offspring_selection_strategy, 100, backup)

    backup_files = glob("backups/evo1/*.population")
    assert len(backup_files) == 20

    # Clear after save
    rmtree("backups", ignore_errors=True)


def test_load_population_backup():
    number_of_genotypes = 40
    input_nodes = 4
    hidden_layers_nodes = []
    output_nodes = 3
    number_of_nn_weights = get_number_of_nn_weights(input_nodes, hidden_layers_nodes, output_nodes)
    weight_lower_threshold = -1
    weight_upper_threshold = 1
    mutation_mean = 0
    mutation_standard_deviation = 0.2

    def phenotype_strategy(genotype):
        return Phenotype.get_phenotype_from_genotype(genotype, input_nodes, hidden_layers_nodes, output_nodes)

    # Fake evaluation function, assign random number form 0 to 1 to each phenotype
    def evaluation_strategy(phenotypes):
        number_of_phenotypes = len(phenotypes)

        evaluation = [0] * number_of_phenotypes
        for index in range(number_of_phenotypes):
            evaluation[index] = random()

        return evaluation

    def parent_selection_strategy(phenotypes_values):
        return get_two_size_tournament_parent_selection(phenotypes_values, number_of_genotypes)

    def mutation_strategy(genotype):
        return SimpleGenotype.get_mutated_genotype(genotype, mutation_mean, mutation_standard_deviation)

    def offspring_selection_strategy(parents, mutated_parents):
        return get_age_based_offspring_selection(parents, mutated_parents)

    initial_population = SimpleGenotype.get_random_genotypes(number_of_genotypes, number_of_nn_weights,
                                                             weight_lower_threshold, weight_upper_threshold)

    backup = [5, "backups/evo1", ".population"]
    evolved_population = get_evolved_population(initial_population, phenotype_strategy, evaluation_strategy,
                                                parent_selection_strategy, mutation_strategy,
                                                offspring_selection_strategy, 100, backup)

    population = handle_backup_load("backups/evo1", '.population')
    assert len(population) == 40
    assert population == evolved_population

    # Clear after save
    rmtree("backups", ignore_errors=True)



from femos.genotypes import SimpleGenotype, UncorrelatedOneStepSizeGenotype, UncorrelatedNStepSizeGenotype


def test_simple_genotype_get_random_genotype():
    number_of_nn_weights = 20
    weight_lower_threshold = -1
    weight_upper_threshold = 1

    random_simple_genotype = SimpleGenotype.get_random_genotype(number_of_nn_weights, weight_lower_threshold,
                                                                weight_upper_threshold)

    assert type(random_simple_genotype) is SimpleGenotype
    assert len(random_simple_genotype.weights) == 20
    for element in random_simple_genotype.weights:
        assert weight_lower_threshold <= element <= weight_upper_threshold


def test_simple_genotype_get_random_genotypes():
    number_of_genotypes = 20

    random_simple_genotypes = SimpleGenotype.get_random_genotypes(number_of_genotypes, 20, -1, 1)
    assert len(random_simple_genotypes) == number_of_genotypes
    for element in random_simple_genotypes:
        assert type(element) is SimpleGenotype


def test_simple_genotype_get_mutated_genotype():
    number_of_nn_weights = 20
    weight_lower_threshold = -1
    weight_upper_threshold = 1
    mutation_mean = 0
    mutation_standard_deviation = 0.2

    random_simple_genotype = SimpleGenotype.get_random_genotype(number_of_nn_weights, weight_lower_threshold,
                                                                weight_upper_threshold)
    mutated_simple_genotype = SimpleGenotype.get_mutated_genotype(random_simple_genotype, 0, 0.3)

    assert random_simple_genotype != mutated_simple_genotype
    assert len(mutated_simple_genotype.weights) == number_of_nn_weights

    for index in range(number_of_nn_weights):
        assert random_simple_genotype.weights[index] != mutated_simple_genotype.weights[index]


def test_uncorrelated_one_step_size_genotype_get_random_genotype():
    number_of_nn_weights = 20
    weight_lower_threshold = -1
    weight_upper_threshold = 1
    mutation_step_size_lower_threshold = -0.3
    mutation_step_size_upper_threshold = 0.3

    random_uoss_genotype = UncorrelatedOneStepSizeGenotype.get_random_genotype(number_of_nn_weights,
                                                                               weight_lower_threshold,
                                                                               weight_upper_threshold,
                                                                               mutation_step_size_lower_threshold,
                                                                               mutation_step_size_upper_threshold)

    assert type(random_uoss_genotype) is UncorrelatedOneStepSizeGenotype
    assert len(random_uoss_genotype.weights) == number_of_nn_weights
    for element in random_uoss_genotype.weights:
        assert weight_lower_threshold <= element <= weight_upper_threshold
    assert random_uoss_genotype.mutation_step_size >= mutation_step_size_lower_threshold
    assert random_uoss_genotype.mutation_step_size <= mutation_step_size_upper_threshold


def test_uncorrelated_one_step_size_genotype_get_random_genotypes():
    number_of_genotypes = 20
    number_of_nn_weights = 20
    weight_lower_threshold = -1
    weights_upper_threshold = 1
    mutation_step_size_lower_threshold = -0.3
    mutation_step_size_upper_threshold = 0.3

    random_uoss_genotypes = UncorrelatedOneStepSizeGenotype.get_random_genotypes(number_of_genotypes,
                                                                                 number_of_nn_weights,
                                                                                 weight_lower_threshold,
                                                                                 weights_upper_threshold,
                                                                                 mutation_step_size_lower_threshold,
                                                                                 mutation_step_size_upper_threshold)

    assert len(random_uoss_genotypes) == number_of_genotypes
    for element in random_uoss_genotypes:
        assert type(element) is UncorrelatedOneStepSizeGenotype


def test_uncorrelated_one_step_size_genotype_mutation():
    number_of_nn_weights = 20
    weight_lower_threshold = -1
    weights_upper_threshold = 1
    mutation_step_size_lower_threshold = -0.3
    mutation_step_size_upper_threshold = 0.3
    mutation_tau1 = 0.01

    random_uoss_genotype = UncorrelatedOneStepSizeGenotype.get_random_genotype(number_of_nn_weights,
                                                                               weight_lower_threshold,
                                                                               weights_upper_threshold,
                                                                               mutation_step_size_lower_threshold,
                                                                               mutation_step_size_upper_threshold)

    mutated_random_uoss_genotype = UncorrelatedOneStepSizeGenotype.get_mutated_genotype(random_uoss_genotype,
                                                                                        mutation_tau1)

    assert random_uoss_genotype != mutated_random_uoss_genotype
    assert len(mutated_random_uoss_genotype.weights) == number_of_nn_weights

    for index in range(number_of_nn_weights):
        assert random_uoss_genotype.weights[index] != mutated_random_uoss_genotype.weights[index]


def test_uncorrelated_n_step_size_genotype_get_random_genotype():
    number_of_nn_weights = 20
    weight_lower_threshold = -1
    weight_upper_threshold = 1
    mutation_step_size_lower_threshold = -0.3
    mutation_step_size_upper_threshold = 0.3

    random_unss_genotype = UncorrelatedNStepSizeGenotype.get_random_genotype(number_of_nn_weights,
                                                                             weight_lower_threshold,
                                                                             weight_upper_threshold,
                                                                             mutation_step_size_lower_threshold,
                                                                             mutation_step_size_upper_threshold)

    assert type(random_unss_genotype) is UncorrelatedNStepSizeGenotype
    assert len(random_unss_genotype.weights) == number_of_nn_weights
    assert len(random_unss_genotype.mutation_step_sizes) == number_of_nn_weights

    for element in random_unss_genotype.weights:
        assert weight_lower_threshold <= element <= weight_upper_threshold

    for element in random_unss_genotype.mutation_step_sizes:
        assert mutation_step_size_lower_threshold <= element <= mutation_step_size_upper_threshold


def test_uncorrelated_n_step_size_genotype_get_random_genotypes():
    number_of_genotypes = 20
    number_of_nn_weights = 20
    weight_lower_threshold = -1
    weight_upper_threshold = 1
    mutation_step_size_lower_threshold = -0.3
    mutation_step_size_upper_threshold = 0.3

    random_unss_genotypes = UncorrelatedNStepSizeGenotype.get_random_genotypes(number_of_genotypes,
                                                                               number_of_nn_weights,
                                                                               weight_lower_threshold,
                                                                               weight_upper_threshold,
                                                                               mutation_step_size_lower_threshold,
                                                                               mutation_step_size_upper_threshold)

    assert len(random_unss_genotypes) == number_of_genotypes
    for element in random_unss_genotypes:
        assert type(element) is UncorrelatedNStepSizeGenotype


def test_uncorrelated_n_step_size_genotype_mutation():
    number_of_nn_weights = 20
    weight_lower_threshold = -1
    weight_upper_threshold = 1
    mutation_step_size_lower_threshold = -0.3
    mutation_step_size_upper_threshold = 0.3
    mutation_tau1 = 0.01
    mutation_tau2 = 0.1

    random_unss_genotype = UncorrelatedNStepSizeGenotype.get_random_genotype(number_of_nn_weights,
                                                                             weight_lower_threshold,
                                                                             weight_upper_threshold,
                                                                             mutation_step_size_lower_threshold,
                                                                             mutation_step_size_upper_threshold)

    mutated_random_unss_genotype = UncorrelatedNStepSizeGenotype.get_mutated_genotype(random_unss_genotype,
                                                                                      mutation_tau1, mutation_tau2)

    assert random_unss_genotype != mutated_random_unss_genotype
    assert len(mutated_random_unss_genotype.weights) == number_of_nn_weights

    for index in range(number_of_nn_weights):
        assert random_unss_genotype.weights[index] != mutated_random_unss_genotype.weights[index]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="femos",
    version="0.0.11",
    author="estissy",
    author_email="estissy@gmail.com",
    description="My small library for neuroevolution",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/estissy/femos",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy',
        'humanize'
    ],
    python_requirements=">=3.0",
)



from enum import Enum
from glob import glob
from os import path, makedirs
from pickle import dump, HIGHEST_PROTOCOL, load
from random import uniform
from statistics import mean, stdev
from time import time
from datetime import datetime, timedelta


class Summary(Enum):
    EPOCH = 1,
    MEAN = 2,
    STDDEV = 3,
    POPULATION_SIZE = 4,
    DURATION = 5,


def get_random_numbers(quantity, lower_threshold, upper_threshold):
    numbers = []
    for index in range(quantity):
        numbers.append(uniform(lower_threshold, upper_threshold))

    return numbers


def get_number_of_nn_weights(input_nodes, hidden_layers_nodes, output_nodes, use_bias=False):
    grouped_nodes = [input_nodes] + hidden_layers_nodes + [output_nodes]

    total = 0
    for index in range(len(grouped_nodes) - 1):
        total += grouped_nodes[index] * grouped_nodes[index + 1]

    if use_bias:
        number_of_bias_weights = sum(hidden_layers_nodes) + output_nodes
        total += number_of_bias_weights

    return total


def get_next_population(population, phenotype_strategy, evaluation_strategy, parent_selection_strategy,
                        mutation_strategy, offspring_selection_strategy):
    start_time = time()
    phenotypes = map(phenotype_strategy, population)
    phenotypes_values = evaluation_strategy(list(phenotypes))

    parent_indices = parent_selection_strategy(phenotypes_values)
    parents = map(lambda parent_index: population[parent_index], parent_indices)

    mutated_parents = map(mutation_strategy, parents)
    offspring = offspring_selection_strategy(population, list(mutated_parents))
    end_time = time()

    return offspring, phenotypes_values, start_time, end_time


def get_population_file_name(extension=".population"):
    current_timestamp = str(int(time()))
    return current_timestamp + extension


def handle_backup(backup_strategy, current_epoch, population):
    if backup_strategy is not None:
        interval = backup_strategy[0]

        if current_epoch % interval == 0:
            backup_directory = backup_strategy[1]
            file_extension = backup_strategy[2]

            makedirs(backup_directory, exist_ok=True)

            file_name = get_population_file_name(file_extension)
            backup_path = path.join(backup_directory, file_name)

            with open(backup_path, "wb+") as dump_file:
                dump(population, dump_file, HIGHEST_PROTOCOL)


def handle_backup_load(backup_path, extension):
    glob_search_string = path.join(backup_path, "*{}".format(extension))
    backup_files = glob(glob_search_string)
    sorted_backup_files = sorted(backup_files)
    last_backup_file = sorted_backup_files[-1]

    with open(last_backup_file, "rb") as loaded_file:
        population = load(loaded_file)

    return population


def get_epoch_summary(summary, epoch, phenotype_values, start_time, end_time):
    results = {}

    if Summary.EPOCH in summary:
        results.update({Summary.EPOCH: epoch})

    if Summary.MEAN in summary:
        results.update({Summary.MEAN: mean(phenotype_values)})

    if Summary.STDDEV in summary:
        results.update({Summary.STDDEV: stdev(phenotype_values)})

    if Summary.POPULATION_SIZE in summary:
        population_size = len(phenotype_values)
        results.update({Summary.POPULATION_SIZE: population_size})

    if Summary.DURATION in summary:
        duration = end_time - start_time
        results.update({Summary.DURATION: duration})

    return results


def handle_epoch_summary(summary_strategy, epoch, phenotype_values, start_time, end_time):
    if summary_strategy is not None:
        interval = summary_strategy[1]

        if epoch % interval == 0:
            summary = summary_strategy[0]
            output = []
            epoch_summary = get_epoch_summary(summary, epoch, phenotype_values, start_time, end_time)

            if Summary.EPOCH in summary:
                output.append(str.format('Epoch: {}', epoch_summary[Summary.EPOCH]))

            if Summary.MEAN in summary:
                output.append(str.format('Mean: {}', epoch_summary[Summary.MEAN]))

            if Summary.STDDEV in summary:
                output.append(str.format('Stddev: {}', epoch_summary[Summary.STDDEV]))

            if Summary.POPULATION_SIZE in summary:
                output.append(str.format('Population size: {}', epoch_summary[Summary.POPULATION_SIZE]))

            if Summary.DURATION in summary:
                output.append(str.format("Duration: {}", epoch_summary[Summary.DURATION]))

            print(' | '.join(output))


def get_end_datetime(duration):
    start_datetime = datetime.now()
    duration_timedelta = timedelta(hours=duration)
    return start_datetime + duration_timedelta


def get_evolved_population(initial_population, phenotype_strategy, evaluation_strategy, parent_selection_strategy,
                           mutation_strategy, offspring_selection_strategy, duration, backup_strategy=None,
                           epoch_summary_strategy=None):
    tmp_population = initial_population
    end_datetime = get_end_datetime(duration)
    epoch = 0

    while datetime.now() <= end_datetime:
        tmp_population, phenotype_values, start_time, end_time = get_next_population(tmp_population, phenotype_strategy,
                                                                                     evaluation_strategy,
                                                                                     parent_selection_strategy,
                                                                                     mutation_strategy,
                                                                                     offspring_selection_strategy)

        epoch += 1
        handle_backup(backup_strategy, epoch, tmp_population)
        handle_epoch_summary(epoch_summary_strategy, epoch, phenotype_values, start_time, end_time)

    return tmp_population
